#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <time.h>

#include "core.h"
#include "socket.h"
#include "timer.h"

#define maxMacAddr 8

#define MAJOR_VERSION (int)0
#define MINOR_VERSION (int)7

#ifndef __packed
#define __packed __attribute__((packed))
#endif

struct USER_INFO {
	uint32_t last;
	char szNickName[30];
	unsigned char dummy1[2];
	uint32_t majorVersion;
	uint32_t minorVersion;
	char szSSID[33];
	char szUID[33];
	unsigned char dummy2[2];
	uint32_t pid;
	uint32_t pspCount;
	uint32_t ping;
} __packed;

struct MAC_ADDRESS {
	unsigned char addr[6];
	unsigned char dummy1[2];
	uint32_t active;
} __packed;

struct DATA_HEADER {
	unsigned char dtype;
	unsigned char doption;
	uint16_t dsize;
	uint32_t sendCount;
	uint32_t recvCount;
} __packed;

struct client_session_data {
	struct {
		int server_connect: 1;
	} state;
	long sendCount;
	long recvCount;

	int ping_highreso;
	struct MAC_ADDRESS mac[maxMacAddr];
	struct USER_INFO info;
};

int listen_port;
unsigned char eof[4]="<E>";

unsigned int swapw(unsigned int a)
{
	return ((a&0x00ff)<<8) + ((a&0xff00)>>8);
}
unsigned int swapl(unsigned int a)
{
	return ((a&0x000000ff)<<24) + ((a&0x0000ff00)<<8) + ((a&0x00ff0000)>>8) + ((a&0xff000000)>>24);
}

#if BYTE_ORDER == BIG_ENDIAN
#define c2sw(a) swapw(a)
#define c2sl(a) swapl(a)
#define s2cw(a) swapw(a)
#define s2cl(a) swapl(a)
#else
#define c2sw(a) (a)
#define c2sl(a) (a)
#define s2cw(a) (a)
#define s2cl(a) (a)
#endif

int set_count_eof(int fd)
{
	struct client_session_data *csd;

	csd = session[fd]->session_data;
	csd->sendCount++;
	WFIFOL(fd,4) = s2cl(csd->sendCount);
	WFIFOL(fd,8) = s2cl(csd->recvCount);
	memcpy(WFIFOP(fd,sizeof(struct DATA_HEADER)+c2sw(WFIFOW(fd,2))),eof,sizeof(eof));
//	printf("send set %d %c %d %d %d %d :",fd,WFIFOB(fd,0),WFIFOB(fd,1),c2sw(WFIFOW(fd,2)),c2sl(WFIFOL(fd,4)),c2sl(WFIFOL(fd,8)));
//	for(i=0;i<c2sw(WFIFOW(fd,2));i++)
//		printf(" %02x",WFIFOB(fd,sizeof(struct DATA_HEADER)+i));
//	printf("\n");

	return 0;
}

int parse_cl(int fd)
{
	int cmd,len;
	struct client_session_data *csd;
	int i;
	struct DATA_HEADER dh;
	struct USER_INFO *ui;
	unsigned char *data;
	int ping;
	int count;
	char commandType[30];
	char tellTo[30];
	char text[4096];
	const unsigned char bc[6] = {0xff,0xff,0xff,0xff,0xff,0xff};

	if(session[fd]->session_data==NULL){
		session[fd]->session_data = calloc(sizeof(struct client_session_data),1);
		if(session[fd]->session_data == NULL){
			printf("parse_cl : out of memory\n");
			exit(1);
		}
	}
	csd = session[fd]->session_data;

	// ڑ؂Ă̂Ōn
	if(session[fd]->eof){
		close(fd);
		delete_session(fd);
		return 0;
	}

	while(RFIFOREST(fd)>=sizeof(struct DATA_HEADER)){
		len=sizeof(struct DATA_HEADER)+c2sw(RFIFOW(fd,2))+sizeof(eof);
		// debug
//		printf("%d %c %d %d %d %d ",fd,RFIFOB(fd,0),RFIFOB(fd,1),c2sw(RFIFOW(fd,2)),c2sl(RFIFOL(fd,4)),c2sl(RFIFOL(fd,8)));
//		for(i=sizeof(struct DATA_HEADER);i<len;i++){
//			printf("%02x ",RFIFOB(fd,i));
//		}
//		printf("\n");
		if(RFIFOREST(fd)<len)
			return 0;
		if(memcmp(RFIFOP(fd,sizeof(struct DATA_HEADER)+c2sw(RFIFOW(fd,2))),eof,sizeof(eof))){
			printf("eof not found\n");
			RFIFOSKIP(fd,RFIFOREST(fd));
			return 0;
		}
		csd->recvCount++;
		data = RFIFOP(fd,sizeof(struct DATA_HEADER));

		memcpy(&dh,RFIFOP(fd,0),sizeof(dh));

		switch(cmd=RFIFOB(fd,0)){
		case 'C':
		case 'U':
			dh.doption = fd;
			dh.dtype = cmd=='C' ? 'c' : 'u' ;
			for(i=0;i<fd_max;i++){
				if(session[i] && session[i]->session_data){
					memcpy(WFIFOP(i,0),&dh,sizeof(dh));
					memcpy(WFIFOP(i,sizeof(dh)),data,c2sw(dh.dsize));
					set_count_eof(i);
					WFIFOSET(i,sizeof(dh)+c2sw(dh.dsize)+sizeof(eof));
				}
			}
			break;

		case 'D':
			data[c2sw(dh.dsize)] = 0;
			sscanf(data, "/%s %s %4096s", commandType, tellTo, text);
			dh.doption = fd;
			dh.dtype = 'c';
			for(i=0;i<fd_max;i++){
				struct client_session_data *csd2;
				if(session[i] && (csd2=session[i]->session_data) && 
				   strcmp(csd2->info.szNickName, tellTo)==0 ){
					snprintf(WFIFOP(fd,sizeof(dh)),4096, "%s << %s", tellTo, text);
					dh.dsize = s2cw(strlen(WFIFOP(fd,sizeof(dh))));
					memcpy(WFIFOP(fd,0),&dh,sizeof(dh));
					set_count_eof(fd);
					WFIFOSET(fd,sizeof(dh)+c2sw(dh.dsize)+sizeof(eof));

					snprintf(WFIFOP(i,sizeof(dh)),4096, "%s >> %s", csd->info.szNickName, text);
					dh.dsize = s2cw(strlen(WFIFOP(i,sizeof(dh))));
					memcpy(WFIFOP(i,0),&dh,sizeof(dh));
					set_count_eof(i);
					WFIFOSET(i,sizeof(dh)+c2sw(dh.dsize)+sizeof(eof));
				}
			}
			break;
		case 'Y':
			data[c2sw(dh.dsize)] = 0;
			sscanf(data, "/%s %4096s", commandType, text);
			dh.doption = fd;
			dh.dtype = 'c';
			for(i=0;i<fd_max;i++){
				struct client_session_data *csd2;
				if(session[i] && (csd2=session[i]->session_data) && 
				   strcmp(csd2->info.szSSID, csd->info.szSSID)==0){
					snprintf(WFIFOP(i,sizeof(dh)),4096, "(%s) %s", csd->info.szNickName, text);
					dh.dsize = s2cw(strlen(WFIFOP(i,sizeof(dh))));
					memcpy(WFIFOP(i,0),&dh,sizeof(dh));
					set_count_eof(i);
					WFIFOSET(i,sizeof(dh)+c2sw(dh.dsize)+sizeof(eof));
				}
			}
			break;

		case 'T':
			dh.doption = fd;
			dh.dtype = 't';
			for(i=0;i<maxMacAddr;i++){
				if (csd->mac[i].active && memcmp(csd->mac[i].addr, &data[6], 6) == 0)
					break;
			}
			if(i==maxMacAddr){
				for(i=0;i<maxMacAddr;i++){
					if(!csd->mac[i].active){
						memcpy(csd->mac[i].addr,&data[6],6);
						csd->mac[i].active=s2cl(1);
						break;
					}
				}
			}
			for(i=0;i<fd_max;i++){
				struct client_session_data *csd2;
				if(i!=fd && // dont send self
				   session[i] && (csd2=session[i]->session_data) && 
				   strcmp(csd2->info.szSSID, csd->info.szSSID)==0){
					int s=0;
					if(memcmp(bc,data,6)==0)
						s=1;
					else {
						int j;
						for(j=0;j<maxMacAddr;j++){
							if(memcmp(csd2->mac[j].addr,data,6)==0){
								s=1;
								break;
							}
						}
					}
					if(s){
						memcpy(WFIFOP(i,0),&dh,sizeof(dh));
						memcpy(WFIFOP(i,sizeof(dh)),data,c2sw(dh.dsize));
						set_count_eof(i);
						WFIFOSET(i,sizeof(dh)+c2sw(dh.dsize)+sizeof(eof));
					}
				}
			}
			break;

		case 'S':
		case 'I':
			ping = c2sl(csd->info.ping);
			memcpy(&csd->info,data,sizeof(struct USER_INFO));
			csd->info.ping = s2cl(ping);
			if (c2sl(csd->info.majorVersion) != MAJOR_VERSION || c2sl(csd->info.minorVersion) != MINOR_VERSION) {
				dh.dtype = 'c';
				snprintf(WFIFOP(fd,sizeof(dh)),512, "o[Wv܂: T[o[Ver%d.%d NCAgVer%d.%d",
						 MAJOR_VERSION, MINOR_VERSION, c2sl(csd->info.majorVersion), c2sl(csd->info.minorVersion));
				dh.dsize = s2cw(strlen(WFIFOP(fd,sizeof(dh))));
				memcpy(WFIFOP(fd,0),&dh,sizeof(dh));
				set_count_eof(fd);
				WFIFOSET(fd,sizeof(dh)+c2sw(dh.dsize)+sizeof(eof));
			}
			if(cmd=='I'){
				i= dh.doption;
				dh.dtype = 'i';
				dh.doption = fd;
				if(session[i] && session[i]->session_data){
					memcpy(WFIFOP(i,0),&dh,sizeof(dh));
					memcpy(WFIFOP(i,sizeof(dh)),data,c2sw(dh.dsize));
					set_count_eof(i);
					WFIFOSET(i,sizeof(dh)+c2sw(dh.dsize)+sizeof(eof));
				}
				break;
			}
			// 'S'
			for(i=0,count=0;i<fd_max;i++){
				if(session[i] && session[i]->session_data){
					count++;
				}
			}
			ui=malloc(sizeof(*ui)*count);
			for(i=0,count=0;i<fd_max;i++){
				struct client_session_data *csd2;
				if(session[i] && (csd2=session[i]->session_data)){
					memcpy(ui+count,&csd2->info,sizeof(*ui));
					count++;
				}
			}
			dh.dsize=s2cw(count*sizeof(*ui));
			for(i=0;i<fd_max;i++){
				if(session[i] && session[i]->session_data){
					WFIFOB(i,0)='a';
					WFIFOB(i,1)=count;
					WFIFOW(i,2)=dh.dsize;
					memcpy(WFIFOP(i,sizeof(dh)),ui,c2sw(dh.dsize));
					set_count_eof(i);
					WFIFOSET(i,sizeof(dh)+c2sw(dh.dsize)+sizeof(eof));
				}
			}
			free(ui);
			break;
		case 'P':	// ping
			memcpy(WFIFOP(fd,0),RFIFOP(fd,0),len);
			WFIFOB(fd,0)='p';
			set_count_eof(fd);
			WFIFOSET(fd,len);
			break;

		case 'p':	// pong
// printf("%d %d ->",fd,csd->ping_highreso);
			csd->ping_highreso = (csd->ping_highreso*3 +(gettick_highreso()- RFIFOL(fd,12)) + 2) / 4 ;
// printf(" %d(%d)\n",csd->ping_highreso,gettick_highreso()-RFIFOL(fd,12));
			csd->info.ping = s2cl(csd->ping_highreso/1000);
		case 'A':	// request 'a'
			for(i=0,count=0;i<fd_max;i++){
				if(session[i] && session[i]->session_data){
					count++;
				}
			}
			ui=malloc(sizeof(*ui)*count);
			for(i=0,count=0;i<fd_max;i++){
				struct client_session_data *csd2;
				if(session[i] && (csd2=session[i]->session_data)){
					memcpy(ui+count,&csd2->info,sizeof(*ui));
					ui[count].ping = s2cl((csd->ping_highreso + csd2->ping_highreso)/1000);
					count++;
				}
			}
			dh.dsize=s2cw(count*sizeof(*ui));

			WFIFOB(fd,0)='a';
			WFIFOB(fd,1)=count;
			WFIFOW(fd,2)=dh.dsize;
			memcpy(WFIFOP(fd,sizeof(dh)),ui,c2sw(dh.dsize));
			set_count_eof(fd);
			WFIFOSET(fd,sizeof(dh)+c2sw(dh.dsize)+sizeof(eof));

			free(ui);

			break;

		default:
			break;
		}
		RFIFOSKIP(fd,len);
	}

	return 0;
}

int send_ping(void)
{
	int i;

	for(i=0;i<fd_max;i++){
		if(session[i] && session[i]->session_data){
			WFIFOB(i,0)='P';
			WFIFOB(i,1)=0;
			WFIFOW(i,2)=s2cw(4);
			WFIFOL(i,12)=gettick_highreso();
			set_count_eof(i);
			WFIFOSET(i,12+4+4);
		}
	}
	return 0;
}

int do_init(int argc,char *argv[])
{
	if(argc<2){
		printf("%s <listen port>\n",argv[0]);
		exit(1);
	}
	listen_port = atoi(argv[1]);

	if(make_listen_port(listen_port)==0){
		printf("make_listen_port %d\n",listen_port);
		exit(1);
	}

	set_defaultparse(parse_cl);
	add_timer_interval(gettick()+10,send_ping,0,0,2*1000);

	return 0;
}
